/* G4LTL: Games for LTL Synthesis
 *
 * Copyright (c) 2013, Chih-Hong Cheng, fortiss (An-Institut der TU Muenchen)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the <organization> nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Chih-Hong Cheng BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package g4ltl;

import g4ltl.utility.ProblemDescription;
import g4ltl.utility.SynthesisEngine;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import javax.swing.ImageIcon;

/**
 * APIs for all utility functions including file parsing and external invocation
 * of G4LTL without using the GUI.
 * 
 * @author Chihhong Cheng
 * @version 0.3 2013/02/28
 */
public class SolverUtility {

     static ImageIcon G4LTL_ICON = new ImageIcon("G4LTL.png");
     static String LTL = "LTL";
     static String INPUT = "INPUT";
     static String OUTPUT = "OUTPUT";
     static String ASSUME = "ASSUME";

     public static ImageIcon getIcon(){
         return G4LTL_ICON;
     }
     
    static String changeSpecToInternalFormat(String spec) {
        spec = spec.replaceAll("ALWAYS", " [] ");
        spec = spec.replaceAll("EVENTUALLY", " <> ");
        spec = spec.replaceAll("UNTIL", " U ");
        spec = spec.replaceAll("NEXT", " X ");
        return spec;
    }

    /**
     * Read from the file and generate the input signals, output signals, and
     * the LTL specification.
     *
     * @param selectedFile the specification file
     * @return input, output, and LTL specification
     * @throws Exception exceptions during the file accessing process.
     */
    static HashMap<String, String> getLTLSpecificationFromFile(File selectedFile) throws Exception {
        HashMap<String, String> map = new HashMap<String, String>();

        StringBuilder ltlSpec = new StringBuilder("");
        StringBuilder inputSpec = new StringBuilder("");
        StringBuilder outputSpec = new StringBuilder("");

        FileReader fr = new FileReader(selectedFile);
        BufferedReader br = new BufferedReader(fr);
        String line;
        while ((line = br.readLine()) != null) {
            if (!line.trim().equalsIgnoreCase("")) {
                if (line.trim().startsWith(INPUT)) {
                    if (inputSpec.toString().equals("")) {
                        inputSpec.append(line.trim().substring(INPUT.length()));
                    } else {
                        inputSpec.append(" , ").append(line.trim().substring(INPUT.length()));
                    }
                } else if (line.trim().startsWith(OUTPUT)) {
                    if (outputSpec.toString().equals("")) {
                        outputSpec.append(line.trim().substring(OUTPUT.length()));
                    } else {
                        outputSpec.append(" , ").append(line.trim().substring(OUTPUT.length()));
                    }

                } else {
                    ltlSpec.append(line.trim()).append("\n");
                }

            }
        }
        br.close();

        map.put(INPUT, inputSpec.toString().trim());
        map.put(OUTPUT, outputSpec.toString().trim());
        map.put(LTL, ltlSpec.toString());
        return map;

    }

    static ArrayList<String> getSignals(String signalText) {
        ArrayList<String> result = new ArrayList<String>();
        String[] signals = signalText.split(",");
        for (String signal : signals) {
            if (signal.trim().equals("") == false) {
                result.add(signal.trim());
            }
        }
        return result;
    }

    static ArrayList<ProblemDescription> parseCompositionalLTLspecification(String ltlSpec, int unrollSteps) {
        ArrayList<ProblemDescription> specs = new ArrayList<ProblemDescription>();

        String[] lines = ltlSpec.split("\\n");
        ProblemDescription pd = new ProblemDescription();
        StringBuilder resultLTL = new StringBuilder("");
        boolean firstLine = true;
        for (String line : lines) {
            if (line.trim().startsWith("!--")) {

                // End of the sub-specification. 
                // Extract the input and output variables (from the line) used in this partial specification.
                String[] str = line.split("[{}]");
                String[] inputSignals = str[1].split(",");
                String[] outputSignals = str[3].split(",");
                for (String input : inputSignals) {
                    pd.getInputSignals().add(input.trim());
                }

                for (String output : outputSignals) {
                    pd.getOutputSignals().add(output.trim());
                }

                // Add the partial specification to the system
                pd.setLtlSpecification(resultLTL.toString());
                pd.setUnrollSteps(unrollSteps);
                pd.setPartialSpecification(true);

                // Add the partial specification into analysis
                specs.add(pd);

                // Reset for a new sub-specification
                pd = new ProblemDescription();
                resultLTL = new StringBuilder("");
                firstLine = true;
            } else {
                // Add the original specification
                if (line.trim().startsWith("##") == false && line.trim().startsWith("!--") == false) {
                    // Cover the line to the spec
                    if (firstLine) {
                        resultLTL.append("(").append(line.trim()).append(")");
                        firstLine = false;
                    } else {
                        resultLTL.append(" && (").append(line.trim()).append(")");
                    }
                }
            }
        }

        return specs;
    }

    /**
     * Generate the LTL formula from the specification, where logical implications
     * are used to replace assumptions stated in the spec.
     * 
     * @param ltlSpec LTL specification under analysis
     * @return an LTL formula which can be understood by ltlbuchi or ltl2ba
     */
    static String parseLTLspecification(String ltlSpec) {
        String[] lines = ltlSpec.split("\\n");
        StringBuilder resultGuarantee = new StringBuilder("");
        StringBuilder resultAssume = new StringBuilder("");
        boolean firstLineGuarantee = true;
        boolean firstLineAssume = true;
        boolean hasAssume = false;
        for (String line : lines) {
            if (line.trim().startsWith("##") == false && line.trim().startsWith("!--") == false) {
                if (line.trim().startsWith(ASSUME)) {
                    hasAssume = true;
                    // Cover the line to the spec
                    if (firstLineAssume) {
                        resultAssume.append("(").append(line.trim().substring(ASSUME.length())).append(")");
                        firstLineAssume = false;
                    } else {
                        resultAssume.append(" && (").append(line.trim().substring(ASSUME.length())).append(")");
                    }
                } else {
                    // Cover the line to the spec
                    if (firstLineGuarantee) {
                        resultGuarantee.append("(").append(line.trim()).append(")");
                        firstLineGuarantee = false;
                    } else {
                        resultGuarantee.append(" && (").append(line.trim()).append(")");
                    }
                }
            }
        }
        if (hasAssume) {
            return "(" + resultAssume.toString() + " ) -> (" + resultGuarantee.toString() + ")";
        } else {
            return resultGuarantee.toString();
        }
    }

    /**
     * Perform LTL synthesis by reading a file.
     * 
     * @param selectedFile File to be parsed. 
     * @param optionTechnique Synthesis algorithm 
     * @param unrollSteps Unroll steps for Co-Buechi solver
     * @param outputFormat Desired output format 
     * @param isStrategyFinding Perform strategy finding (prove existence) or counter-strategy finding (prove non-existence)
     * @return Strategy or error message in textural form.
     * @throws Exception 
     */
    public String synthesizeFromFile(File selectedFile, int optionTechnique, int unrollSteps, int outputFormat, boolean isStrategyFinding) throws Exception {
        
        HashMap<String, String> map = getLTLSpecificationFromFile(selectedFile);

        ProblemDescription prob = new ProblemDescription(
                getSignals(map.get(INPUT)),
                getSignals(map.get(OUTPUT)),
                SolverUtility.parseLTLspecification(changeSpecToInternalFormat(map.get(LTL))),
                unrollSteps, true);

        // Invoke the engine, and redirect the synthesized result to the result panel.
        SynthesisEngine engine = new SynthesisEngine();

        if (optionTechnique == SynthesisEngine.COBUECHI_SOLVER) {
            return engine.invokeMonolithicCoBuechiEngine(prob, true, outputFormat, isStrategyFinding);
        } else {
            return engine.invokeMonolithicBuechiEngine(prob, true, outputFormat, isStrategyFinding);
        }

    }
}
