/* G4LTL: Games for LTL Synthesis
 *
 * Copyright (c) 2013, Chih-Hong Cheng fortiss (An-Institut der TU Muenchen)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the <organization> nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Chih-Hong Cheng BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package g4ltl.utility;

import g4ltl.arena.EdgeElement;
import g4ltl.arena.GameArena;
import g4ltl.arena.VertexEdgeSet;
import g4ltl.gametranslation.cobuechi.CoBuechiSafetyReduction;
import g4ltl.gametranslation.cobuechi.EquivalenceClass;
import g4ltl.utility.mealymachine.MealyMachine;
import g4ltl.utility.mealymachine.MealyMachineEdgeElement;
import gov.nasa.ltl.graph.Edge;
import gov.nasa.ltl.graph.Graph;
import gov.nasa.ltl.graph.Node;
import gov.nasa.ltl.trans.LTL2Buchi;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.*;
import jdd.bdd.BDD;
import jdd.bdd.Permutation;

/**
 * The backend synthesis engine, including a symbolic safety game solver.
 *
 * @author Chihhong Cheng
 * @version 0.2 2012/08/08
 */
public class SynthesisEngine {

    /**
     *  Solver option: Co-Buechi + safety.
     */
    public static int COBUECHI_SOLVER = 0;
    /** 
     * Solver option: Buechi.
     */
    public static int BUECHI_SOLVER = 1;
    /** 
     * Output option: pseudo code format.
     */
    public static int OUTPUT_PSUEDO_CODE = 0;
    /** 
     * Output option: SAL format (SAL is a model checker by SRI International).
     */
    public static int OUTPUT_SAL = 1;
    /** 
     * Output option: Ptolemy II FSMACtor format (Ptolemy II is developed by UC Berkeley).
     */
    public static int OUTPUT_FSM_ACTOR_PTOLEMY = 2;
    /**
     * Maximum number of BDD nodes used in JDD.
     */
    private static final int BDD_MAX_NODE_TABLE_SIZE = 2000000;
    /**
     * Maximum size of cache used in JDD.
     */
    private static final int BDD_MAX_CACHE_SIZE = 400000;
    /**
     * BDD data structure used in the synthesis engine.
     */
    private BDD bdd = new BDD(BDD_MAX_NODE_TABLE_SIZE, BDD_MAX_CACHE_SIZE);
    /**
     * Array that maintains the ordering of variables used in BDD.
     */
    private int[] variableArray;
    /**
     * Data structure to store the index-name pair of each arena.
     */
    private ArrayList<String> vertexNameBuechiAutomaton;
    private ArrayList<String> vertexNameArena;
    /**
     * Maximumly allowed visited final states in Co-Buechi automata.
     */
    private static int MAX_VISIT_COBUECHI_FINAL_STATE = 5;

    /**
     * Solve a Buechi game symbolically and generate a controller (Mealy Machine).
     * 
     * @param gameArena Buechi game in graph form
     * @param finalEnvVertices indices for final vertices
     * @param proveExistence prove existence of strategy or prove non-existence by finding a counter-strategy
     * @return MealyMachine as a strategy (having explicit variable to answer a strategy is found)
     */
    private MealyMachine analyzeBuechiGame(GameArena gameArena,
            ArrayList<Integer> finalEnvVertices, boolean proveExistence) {

        // Step 1-a: Clean up the memory previously used in BDD, and assign new memory for them.
        bdd.cleanup();

        bdd = new BDD(BDD_MAX_NODE_TABLE_SIZE, BDD_MAX_CACHE_SIZE);

        // Step 1-b: Declare the variable for controller strategies.
        int strategy = bdd.getZero();

        // Step 2-a: Decide and declare the number of variables used in the BDD. 

        int totalNumberOfVariables = ((int) (Math.ceil(Math.log10(gameArena.vertexList.size()) / Math.log10(2)))) * 2;
        int NUM_OF_BITS_FOR_STATE = (int) (Math.ceil(Math.log10(gameArena.vertexList.size()) / Math.log10(2)));

        variableArray = new int[totalNumberOfVariables];
        for (int i = 0; i < totalNumberOfVariables; i++) {
            variableArray[i] = bdd.createVar();
        }

        // Step 2-b: Declare and generate the set of plant transitions and controller 
        // transitions with initial value equals FALSE.
        int plantTransition = bdd.getZero();
        int controllerTransition = bdd.getZero();

        for (VertexEdgeSet vertexEdgeSet : gameArena.vertexList) {
            for (EdgeElement edge : vertexEdgeSet.edgeSet) {

                int transition = bdd.getOne();

                char[] sbits = padZeroToString(Integer.toBinaryString(vertexEdgeSet.getVertexID()), NUM_OF_BITS_FOR_STATE).toCharArray();
                for (int j = 0; j < NUM_OF_BITS_FOR_STATE; j++) {
                    if (sbits[j] == '1') {
                        transition = bdd.andTo(transition, variableArray[pre(j)]);
                    } else {
                        transition = bdd.andTo(transition, bdd.not(variableArray[pre(j)]));
                    }
                }

                char[] dbits = padZeroToString(Integer.toBinaryString(edge.getDestVertexID()), NUM_OF_BITS_FOR_STATE).toCharArray();
                for (int j = 0; j < NUM_OF_BITS_FOR_STATE; j++) {
                    if (dbits[j] == '1') {
                        transition = bdd.andTo(transition, variableArray[post(j)]);
                    } else {
                        transition = bdd.andTo(transition, bdd.not(variableArray[post(j)]));
                    }
                }

                if (vertexEdgeSet.getVertexProperty() == VertexEdgeSet.INITIAL_PLANT || vertexEdgeSet.getVertexProperty() == VertexEdgeSet.NONINITIAL_PLANT) {
                    plantTransition = bdd.orTo(plantTransition, transition);
                } else {
                    controllerTransition = bdd.orTo(controllerTransition, transition);
                }
                bdd.deref(transition);
            }
        }

        // Step 2-c: Define variable substitution sequence (permutation)

        int[] p1 = new int[NUM_OF_BITS_FOR_STATE];
        int[] p2 = new int[NUM_OF_BITS_FOR_STATE];
        for (int i = 0; i < NUM_OF_BITS_FOR_STATE; i++) {
            p1[i] = variableArray[pre(i)];
            p2[i] = variableArray[post(i)];
        }


        Permutation perm = bdd.createPermutation(p1, p2);
        Permutation permForward = bdd.createPermutation(p2, p1);

        // Step 2-d: Define the cube to remove x_plum, y_plum, z_plum using the
        // existential quantifier (result only with x, y, z)
        int cube = bdd.getOne();
        int cubeForward = bdd.getOne();
        for (int i = 0; i < NUM_OF_BITS_FOR_STATE; i++) {
            // cube = bdd.andTo(cube, variableArray[START_OF_VERTEX_INDEX_POST + i]);
            cube = bdd.andTo(cube, variableArray[post(i)]);
            cubeForward = bdd.andTo(cubeForward, variableArray[pre(i)]);
        }


        // Step 3-a: Generate the set of initial states: (1, {1})
        // This should be examined by checking all vertices with .
        int initialCondition = bdd.getZero();
        int initialVertexId = -1;
        for (VertexEdgeSet vertexEdgeSet : gameArena.vertexList) {
            if (vertexEdgeSet.getVertexProperty() == VertexEdgeSet.INITIAL_CONTROL || vertexEdgeSet.getVertexProperty() == VertexEdgeSet.INITIAL_PLANT) {
                
                char[] sbits = padZeroToString(Integer.toBinaryString(vertexEdgeSet.getVertexID()), NUM_OF_BITS_FOR_STATE).toCharArray();
                int s = bdd.getOne();
                for (int j = 0; j < NUM_OF_BITS_FOR_STATE; j++) {
                    if (sbits[j] == '1') {
                        s = bdd.andTo(s, variableArray[pre(j)]);
                    } else {
                        s = bdd.andTo(s, bdd.not(variableArray[pre(j)]));
                    }
                }

                initialVertexId = vertexEdgeSet.getVertexID();
                initialCondition = bdd.orTo(initialCondition, s);
                bdd.deref(s);
            }
        }

        // Step 4: Read the specification and invoke suitable verification engine.

        // Step 4-a: Generate the Goal condition: One set of final states.
        int finalStates = bdd.getZero();
        for (Integer integer : finalEnvVertices) {

            char[] sbits = padZeroToString(Integer.toBinaryString(integer.intValue()), NUM_OF_BITS_FOR_STATE).toCharArray();
            int state = bdd.getOne();
            for (int j = 0; j < NUM_OF_BITS_FOR_STATE; j++) {
                if (sbits[j] == '1') {
                    state = bdd.andTo(state, variableArray[pre(j)]);
                } else {
                    state = bdd.andTo(state, bdd.not(variableArray[pre(j)]));
                }
            }

            finalStates = bdd.orTo(finalStates, state);
        }

        // Step 4-b: Invoke the synthesis engine.
        if (proveExistence) {
            strategy = findWinningStrategyBuechi(finalStates, initialCondition, controllerTransition, plantTransition, perm, cube);
            strategy = bdd.andTo(strategy, controllerTransition);
        } else {
            // To prove the non-existence, we need to switch the role of plant and control 
            strategy = findWinningStrategyBuechi(finalStates, initialCondition, plantTransition, controllerTransition, perm, cube);
        }


        // Step 5: Interpret the strategy (and being back to the caller if needed):

        if (proveExistence == false) {
            MealyMachine machine = new MealyMachine();
            if (strategy == bdd.getZero()) {
                // unable to find a counter-witness
                machine.setSolution(false);
            } else {
                // a counter-witness found
                machine.setSolution(true);
            }
            return machine;
        }

        // Prune the strategy by only considering edges which are led from initial states.
        int totalTransition = bdd.ref(bdd.or(strategy, plantTransition));
        int preImage = bdd.ref(initialCondition);
        int postImage = bdd.getZero();
        do {
            postImage = bdd.orTo(postImage, bdd.replace(bdd.relProd(preImage, totalTransition, cubeForward), permForward));
            if (bdd.biimp(preImage, postImage) == bdd.getOne()) {
                break;
            }
            preImage = bdd.ref(postImage);
            bdd.deref(postImage);
        } while (true);
        strategy = bdd.andTo(strategy, postImage);
        bdd.deref(preImage);
        bdd.deref(postImage);
        bdd.deref(totalTransition);

        // Process the strategy to the specified output format
        PrintStream out = System.out;

        ByteArrayOutputStream barray = new ByteArrayOutputStream();
        PrintStream printStreamByteArray = new PrintStream(barray);
        System.setOut(printStreamByteArray);

        bdd.printSet(strategy);
        String strategyStringFormat = barray.toString();

        // Redirect the output stream back to console again.
        System.setOut(out);

        return generateMealyMachineBuechi(strategyStringFormat, initialVertexId, gameArena);

    }

    /**
     * Solve a safety game symbolically and generate a controller (Mealy Machine).
     * 
     * @param safetyArena translated safety game in graph form
     * @param initialVertex initial vertex in the game
     * @param riskVertex risk vertex in the game
     * @param proveExistence prove existence of strategy or prove non-existence by finding a counter-strategy
     * @return MealyMachine as a strategy (having explicit variable to answer a strategy is found)
     */
    private MealyMachine analyzeSafetyGameFromCoBuechi(ArrayList<EquivalenceClass> safetyArena,
            EquivalenceClass initialVertex, EquivalenceClass riskVertex, boolean proveExistence) {

        bdd.cleanup();

        bdd = new BDD(BDD_MAX_NODE_TABLE_SIZE, BDD_MAX_CACHE_SIZE);

        int planttransition = bdd.getZero();
        int controllertransition = bdd.getZero();

        int totalNumberOfVariables = ((int) (Math.ceil(Math.log10(safetyArena.size()) / Math.log10(2)))) * 2;

        int NUM_OF_BITS_FOR_STATE = (int) (Math.ceil(Math.log10(safetyArena.size()) / Math.log10(2)));

        variableArray = new int[totalNumberOfVariables];
        for (int i = 0; i < totalNumberOfVariables; i++) {
            variableArray[i] = bdd.createVar();
        }


        int initialCondition = bdd.getOne();
        char[] bits = padZeroToString(Integer.toBinaryString(initialVertex.id), NUM_OF_BITS_FOR_STATE).toCharArray();
        for (int j = 0; j < NUM_OF_BITS_FOR_STATE; j++) {
            if (bits[j] == '1') {
                initialCondition = bdd.andTo(initialCondition, variableArray[pre(j)]);
            } else {
                initialCondition = bdd.andTo(initialCondition, bdd.not(variableArray[pre(j)]));
            }
        }

        // long startTime = System.currentTimeMillis();

        // Caching the vertex computation to avoid repeated computation.
        int[] stateIdBDDPre = new int[safetyArena.size()];
        int[] stateIdBDDPost = new int[safetyArena.size()];
        for (int i = 0; i < safetyArena.size(); i++) {

            int sPre = bdd.ref(bdd.getOne());
            int sPost = bdd.ref(bdd.getOne());
            char[] sbits = padZeroToString(Integer.toBinaryString(i), NUM_OF_BITS_FOR_STATE).toCharArray();
            for (int j = 0; j < NUM_OF_BITS_FOR_STATE; j++) {
                if (sbits[j] == '1') {
                    sPre = bdd.andTo(sPre, variableArray[pre(j)]);
                    sPost = bdd.andTo(sPost, variableArray[post(j)]);
                } else {
                    sPre = bdd.andTo(sPre, bdd.not(variableArray[pre(j)]));
                    sPost = bdd.andTo(sPost, bdd.not(variableArray[post(j)]));
                }
            }
            stateIdBDDPre[i] = sPre;
            stateIdBDDPost[i] = sPost;
        }
        // long endTime = System.currentTimeMillis();
        //   System.out.println("Total elapsed time in game creation is :" + (endTime - startTime));


        for (EquivalenceClass source : safetyArena) {

            if (source.isEnv == true) {
                for (EquivalenceClass dest : source.successor.values()) {
                    planttransition = bdd.orTo(planttransition, bdd.and(stateIdBDDPre[source.id], stateIdBDDPost[dest.id]));
                }
            } else {
                for (EquivalenceClass dest : source.successor.values()) {
                    controllertransition = bdd.orTo(controllertransition, bdd.and(stateIdBDDPre[source.id], stateIdBDDPost[dest.id]));
                }
            }
        }
        // endTime = System.currentTimeMillis();
        //  System.out.println("Total elapsed time in game creation is :" + (endTime - startTime));


        int[] p1 = new int[NUM_OF_BITS_FOR_STATE];
        int[] p2 = new int[NUM_OF_BITS_FOR_STATE];

        for (int i = 0; i < NUM_OF_BITS_FOR_STATE; i++) {
            p1[i] = variableArray[pre(i)];
            p2[i] = variableArray[post(i)];
        }

        Permutation perm = bdd.createPermutation(p1, p2);

        int cube = bdd.getOne();
        for (int i = 0; i < NUM_OF_BITS_FOR_STATE; i++) {
            cube = bdd.andTo(cube, variableArray[post(i)]);
        }

        int riskStates = bdd.getOne();
        char[] sbits = padZeroToString(Integer.toBinaryString(riskVertex.id), NUM_OF_BITS_FOR_STATE).toCharArray();
        for (int j = 0; j < NUM_OF_BITS_FOR_STATE; j++) {
            if (sbits[j] == '1') {
                riskStates = bdd.andTo(riskStates, variableArray[pre(j)]);
            } else {
                riskStates = bdd.andTo(riskStates, bdd.not(variableArray[pre(j)]));
            }
        }


        // Invoke the synthesis engine.
        // startTime = System.currentTimeMillis();
        int riskStrategy;
        if (proveExistence) {
            riskStrategy = findWinningStrategySafety(riskStates, initialCondition, controllertransition, planttransition, perm, cube);
        } else {
            // For proving non-existence of the controller, we switch the role between environment and control
            riskStrategy = findWinningStrategySafety(riskStates, initialCondition, planttransition, controllertransition, perm, cube);
        }

        // long  endTime = System.currentTimeMillis();
        // System.out.println("Total elapsed time in game solving is :" + (endTime - startTime));     


        if (proveExistence == false) {
            MealyMachine machine = new MealyMachine();
            if (riskStrategy == bdd.getOne()) {
                // No counter-witness exists
                machine.setSolution(false);
            } else {
                // Counter-witness exists
                machine.setSolution(true);
            }
            return machine;
        }

        // Redirect the output stream to string, such that we can interpret the printed BDD set for analysis.
        PrintStream out = System.out;

        ByteArrayOutputStream barray = new ByteArrayOutputStream();
        PrintStream printStreamByteArray = new PrintStream(barray);
        System.setOut(printStreamByteArray);
        bdd.printSet(bdd.and(controllertransition, bdd.not(riskStrategy)));
        String strategyStringFormat = barray.toString();

        // Redirect the output stream back to console again.
        System.setOut(out);

        return generateMealyMachineSafety(strategyStringFormat, initialVertex, safetyArena);

    }

    /**
     * Create the game arena from a Buechi automaton by partitioning input and output signals.
     * 
     * @param inputVariables Input variables
     * @param outputVariables Output variables
     * @param graph Buechi automaton
     * @return the corresponding arena with two players
     */
    private GameArena createGameArena(ArrayList<String> inputVariables, ArrayList<String> outputVariables, Graph graph) {

        // If we need to use aggressive reduction, our heuristic is to select vertices 
        // (other than the initial node) that has the least incoming and outgoing edges.

        if (Debug.DEBUG) {
            Node init = graph.getInit();
            List vertices = graph.getNodes();
            for (Iterator i = vertices.iterator(); i.hasNext();) {
                Node n = (Node) i.next();
                if (init == n) {
                    if (n.getBooleanAttribute("accepting")) {
                        System.out.println("State (init, acc) " + n.getId());
                    } else {
                        System.out.println("State (init) " + n.getId());
                    }
                } else {
                    if (n.getBooleanAttribute("accepting")) {
                        System.out.println("State (acc) " + n.getId());
                    } else {
                        System.out.println("State " + n.getId());
                    }
                }
                for (Iterator j = n.getOutgoingEdges().iterator(); j.hasNext();) {
                    Edge e = (Edge) j.next();
                    System.out.println("\t (" + e.getGuard() + ") -> " + e.getNext().getId());
                }
            }
        }

        ArrayList<Node> vertexList = new ArrayList<Node>();

        // Extract all vertices.      
        List nodes = graph.getNodes();
        for (Iterator i = nodes.iterator(); i.hasNext();) {
            vertexList.add((Node) i.next());
        }

        // Generate all possible bit vectors of the input and outputs
        ArrayList<String> initialVectorList = new ArrayList<String>();
        initialVectorList.add("");
        ArrayList<String> inputBitVectors = generateBitVectors(0, inputVariables.size(), initialVectorList);
        initialVectorList = new ArrayList<String>();
        initialVectorList.add("");
        ArrayList<String> outputBitVectors = generateBitVectors(0, outputVariables.size(), initialVectorList);

        GameArena arena = new GameArena();

        // The ordering of the Buechi automaton follows the index of the
        // vertexList being retrieved. Each index i, it is named in
        // vertexNameArena named "Ei", and stored in the index 
        // i * (size of all possible inputs + 1).       
        vertexNameBuechiAutomaton = new ArrayList<String>();
        // It stores E0, C0_<inputvector1>, ..., E1, C1_<inputvector1>,
        vertexNameArena = new ArrayList<String>();

        // Create the dummy risk vertex which contains a self-loop.


        // Create all environment vertices in the muller arena by extracting from Buechi automata.
        for (Node s : vertexList) {
            vertexNameBuechiAutomaton.add(String.valueOf(s.getId()));
            vertexNameArena.add("E" + String.valueOf(vertexNameBuechiAutomaton.size() - 1));

            VertexEdgeSet vEnv = new VertexEdgeSet();
            int envID = vertexNameArena.size() - 1;
            vEnv.setVertexID(envID);
            if (s.getBooleanAttribute("accepting")) {
                vEnv.setVertexColor(VertexEdgeSet.COLOR_FINAL);
            }
            if (s == graph.getInit()) {
                vEnv.setVertexProperty(VertexEdgeSet.INITIAL_PLANT);
            } else {
                vEnv.setVertexProperty(VertexEdgeSet.NONINITIAL_PLANT);
            }

            // Add the environment vertex
            arena.vertexList.add(vEnv);

            // Create corresponding control vertices with names matching the inputVectors
            for (String vector : inputBitVectors) {
                VertexEdgeSet vCtrl = new VertexEdgeSet();
                vertexNameArena.add("C" + String.valueOf(vertexNameBuechiAutomaton.size() - 1) + "_" + vector);
                vCtrl.setVertexID(vertexNameArena.size() - 1);
                vCtrl.setVertexProperty(VertexEdgeSet.NONINITIAL_CONTROL);
                arena.vertexList.add(vCtrl);
                // Create the environment edge
                EdgeElement edge = new EdgeElement(vEnv.getVertexID(), vCtrl.getVertexID(), vector, 0);
                // Add the edge for the environment                    
                vEnv.edgeSet.add(edge);
            }

        }

        if (Debug.DEBUG) {
            System.out.println("------");
            System.out.println("vertexNameBuechiAutomaton: " + vertexNameBuechiAutomaton.toString());
            System.out.println("vertexNameArena: " + vertexNameArena.toString());
            System.out.println("arena: " + arena.vertexList.size());
        }


        // Create control transitions in the arena: for each edge in the Buechi automaton, 
        // we need to know what is the input (so we can move the the correct control vertex)
        // and what is the output (so we can label the edge with the output).        

        for (Iterator i = nodes.iterator(); i.hasNext();) {
            Node n = (Node) i.next();
            for (Iterator j = n.getOutgoingEdges().iterator(); j.hasNext();) {
                Edge edge = (Edge) j.next();

                ArrayList<String> usedInputBitVectors = new ArrayList<String>();
                ArrayList<String> usedOutputBitVectors = new ArrayList<String>();
                usedInputBitVectors.addAll(inputBitVectors);
                usedOutputBitVectors.addAll(outputBitVectors);

                // Perform normal processing to create edges. The destination is the normal one. 
                String environmentDest = "E" + String.valueOf(vertexNameBuechiAutomaton.indexOf(String.valueOf(edge.getNext().getId())));
                int environmentDestIndex = vertexNameArena.indexOf(environmentDest);

                if (edge.getGuard().equals("-")) {
                    // There is no restriction on the input vectors used.
                } else {
                    StringTokenizer tok = new StringTokenizer(new String(edge.getGuard()), "&");

                    while (tok.hasMoreTokens()) {
                        String token = tok.nextToken();
                        if (token.trim().startsWith("!") && inputVariables.contains(token.trim().substring(1))) {
                            for (String vector : inputBitVectors) {
                                if (vector.substring(inputVariables.indexOf(token.trim().substring(1)),
                                        inputVariables.indexOf(token.trim().substring(1)) + 1).equals("1")) {
                                    // Remove it
                                    usedInputBitVectors.remove(vector);
                                }
                            }
                        } else if (inputVariables.contains(token.trim())) {
                            for (String vector : inputBitVectors) {
                                if (vector.substring(
                                        inputVariables.indexOf(token.trim()), inputVariables.indexOf(token.trim()) + 1).equals("0")) {
                                    // Remove it
                                    usedInputBitVectors.remove(vector);
                                }
                            }
                        } else if (token.trim().startsWith("!") && outputVariables.contains(token.trim().substring(1))) {
                            for (String vector : outputBitVectors) {
                                if (vector.substring(outputVariables.indexOf(token.trim().substring(1)),
                                        outputVariables.indexOf(token.trim().substring(1)) + 1).equals("1")) {
                                    // Remove it
                                    usedOutputBitVectors.remove(vector);
                                }
                            }
                        } else if (outputVariables.contains(token.trim())) {
                            for (String vector : outputBitVectors) {
                                if (vector.substring(
                                        outputVariables.indexOf(token.trim()), outputVariables.indexOf(token.trim()) + 1).equals("0")) {
                                    // Remove it
                                    usedOutputBitVectors.remove(vector);
                                }
                            }
                        }
                    }
                }


                String controlVertexPrefix = "C"
                        + String.valueOf(vertexNameBuechiAutomaton.indexOf(String.valueOf(edge.getSource().getId())));

                for (String input : usedInputBitVectors) {
                    int controlSourceIndex = vertexNameArena.indexOf(controlVertexPrefix + "_" + input);

                    VertexEdgeSet v = arena.vertexList.get(controlSourceIndex);
                    for (String output : usedOutputBitVectors) {
                        // Merge the same destination for the  same output vector. 
                        boolean similarDest = false;
                        for (EdgeElement e : v.edgeSet) {
                            if (e.getDestVertexID()
                                    == environmentDestIndex) { // Append the output to the dest 
                                if (e.getEdgeLabel().contains(output) == false) {
                                    // e.setEdgeLabel(e.getEdgeLabel() + "," + output);
                                    e.getEdgeLabel().add(output);
                                }
                                similarDest = true;
                                break;
                            }
                        }
                        if (similarDest == false) {
                            EdgeElement e = new EdgeElement(controlSourceIndex,
                                    environmentDestIndex, output, 0);
                            v.edgeSet.add(e);
                        }

                    }
                }

            }
        }

        // Printout the translated arena in textural form.
        if (Debug.DEBUG) {
            System.out.println("\n------" + "Buechi/Co-Buechi arena (for output selection)" + "------");
            for (VertexEdgeSet v : arena.vertexList) {
                System.out.println(vertexNameArena.get(v.getVertexID()));
                if (v.getVertexProperty() == VertexEdgeSet.INITIAL_PLANT) {
                    System.out.println("INITIAL");
                }
                for (EdgeElement e : v.edgeSet) {
                    System.out.println("\t" + e.getEdgeLabel() + "-->" + vertexNameArena.get(e.getDestVertexID()));
                }
            }
        }

        return arena;

    }

    /**
     * Find the winning strategy for a Buechi game.
     * 
     * @param finalStates the set of final states
     * @param initialCondition the set of initial states
     * @param controllertransition the set of controller transitions
     * @param planttransition the set of environment transitions
     * @param perm 
     * @param cube 
     * @return index to the bdd for the symbolic startegy
     */
    int findWinningStrategyBuechi(int finalStates, int initialCondition, int controllerTransition, int plantTransition, Permutation perm, int cube) {

        // System.out.println("STEP 1: Calculating recurrence region over final states [Recur(F)]");
        int recurPre = bdd.ref(finalStates);
        int i = 1;
        do {
            int attractorPre = bdd.getZero();
            int j = 1;
            do {
                // [For attractors with controllable locations]
                // Step a1: For ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
                // Step a2: perform AND conjunction with controllertransition to derive set of states
                // Step a3: quantify out x_plum, y_plum, z_plum, and these states are the new attractor ATTR'.

                //int stepa0 = bdd.ref(bdd.exists(bdd.or(attractorPre, recurPre), cube));
                // int stepa1 = bdd.replace(stepa0, perm);
                int stepa1 = bdd.replace(bdd.or(attractorPre, recurPre), perm);
                int stepa2 = bdd.and(stepa1, controllerTransition);
                int stepa3 = bdd.ref(bdd.exists(stepa2, cube));

                int attractorPost_control_state = bdd.ref(stepa3);
                // stream.println("Attractor with control states under processing in step " + j + ":");
                // bdd.printSet(attractorPost_control_state);

                // [For attractors from uncontrollable locations]
                // Step b1: For ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
                // Step b2: For results in step1, perform AND conjunction with planttransition to derive set of states
                // Step b3: quantify out x_plum, y_plum, z_plum, and these states are the new attractor.
                // Step b4: For these states, we "must" make sure that they do not have any transition which leads
                // regions outside the attractor.
                // Step b5: For S\ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
                // Step b6: For results in step5, perform AND conjunction with planttransition to derive set of states
                // Step b7: quantify out x_plum, y_plum, z_plum
                // Step b8: Perform set difference {Step3}\{Step7}, then states in {Step3}\{Step7} will be guaranteed to
                // have EVERY transition leading to the attractor.

                // int stepb0 = bdd.ref(bdd.exists(bdd.or(attractorPre, recurPre), cube));
                // int stepb1 = bdd.ref(bdd.replace(stepb0, perm));
                int stepb1 = bdd.ref(bdd.replace(bdd.or(attractorPre, recurPre), perm));
                int stepb2 = bdd.ref(bdd.and(stepb1, plantTransition));
                int stepb3 = bdd.ref(bdd.exists(stepb2, cube));

                int step = bdd.ref(bdd.not(bdd.or(attractorPre, recurPre))); // S\ATTR
                int stepb5_0 = bdd.ref(bdd.exists(step, cube));
                int stepb5 = bdd.ref(bdd.replace(stepb5_0, perm));
                int stepb6 = bdd.ref(bdd.and(stepb5, plantTransition));
                int stepb7 = bdd.ref(bdd.exists(stepb6, cube));
                int attractorPost_plant_state = bdd.and(stepb3, bdd.not(stepb7));

                // Generate the attractor
                int attractorPost = bdd.ref(bdd.or(attractorPre, bdd.or(attractorPost_control_state, attractorPost_plant_state)));

                // Remove all unused BDD
                // bdd.deref(stepa0);
                bdd.deref(stepa1);
                bdd.deref(stepa2);
                bdd.deref(stepa3);
                bdd.deref(attractorPost_control_state);

                // bdd.deref(stepb0);
                bdd.deref(stepb1);
                bdd.deref(stepb2);
                bdd.deref(stepb3);
                bdd.deref(step);
                bdd.deref(stepb5_0);
                bdd.deref(stepb5);
                bdd.deref(stepb6);
                bdd.deref(stepb7);
                bdd.deref(attractorPost_plant_state);

                // Check if the set of attractors saturates
                if (bdd.ref(bdd.biimp(attractorPre, attractorPost)) == bdd.getOne()) {
                    // System.out.println("------ Finish calculating the attractor (winning region) ----------");
                    // System.out.println("------ Printing the attractor ----------");
                    // bdd.printSet(attractorPost);
                    break;

                } else {
                    j++;
                    bdd.deref(attractorPre); // Dereference the previous image
                    attractorPre =
                            bdd.ref(attractorPost); // Assign the post to be the newly generated image
                    // bdd.printSet(attractorPre);

                    bdd.deref(attractorPost);
                    // System.out.println("-------------------------------------");
                }

            } while (true);

            // Here attractorPre = Attr(recurPre);
            // Intersect with the set of final states
            int recurPost = bdd.ref(bdd.and(finalStates, attractorPre));
            if (bdd.ref(bdd.biimp(recurPre, recurPost)) == bdd.getOne()) {
                // System.out.println("------ Finish calculating the recurrence region ----------");
                // System.out.println("------ Printing the recurrence region ----------");
                // bdd.printSet(bdd.exists(recurPre, cube));
                break;

            } else {
                i++;
                bdd.deref(recurPre); // Dereference the previous image
                recurPre = bdd.ref(recurPost); // Assign the post to be the newly generated image
                bdd.deref(recurPost);
            }

        } while (true);

        if (recurPre == bdd.getZero()) {
            return bdd.getZero();
        }
        // System.out.println("STEP 2: Calculating the attractor over the recurrence region [Attr(Recur(F))]");
        // bdd.printSet(recurPre);

        int attractorPre = bdd.ref(recurPre);
        int controllerSynthesisStrategy = bdd.getZero();

        i = 1;
        do {

            // [For attractors with controllable locations]
            // Step a1: For ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
            // Step a2: perform AND conjunction with controllertransition to derive set of states
            // Step a3: quantify out x_plum, y_plum, z_plum, and these states are the new attractor ATTR'.

            // int stepa0 = bdd.ref(bdd.exists(attractorPre, cube));
            // int stepa1 = bdd.replace(stepa0, perm);
            int stepa1 = bdd.replace(attractorPre, perm);
            int stepa2 = bdd.and(stepa1, controllerTransition);

            if (i == 1) {
                controllerSynthesisStrategy = bdd.orTo(controllerSynthesisStrategy, stepa2);
            } else {
                int stepNew1 = bdd.ref(bdd.and(controllerSynthesisStrategy, stepa2));
                if (bdd.ref(stepNew1) == bdd.getZero()) {
                    System.out.println("GAVS+: There is something wierd here.");
                }
                int stepNew2 = bdd.ref(bdd.exists(stepNew1, cube));
                int stepNew3 = bdd.ref(bdd.and(controllerSynthesisStrategy, bdd.not(stepNew2)));
                int stepNew4 = bdd.ref(bdd.and(stepa2, bdd.not(stepNew2)));
                bdd.deref(controllerSynthesisStrategy);

                controllerSynthesisStrategy =
                        bdd.or(stepNew3, bdd.or(stepNew4, stepNew1));
                // System.out.println("stepNew5");
                // bdd.printSet(stepNew4);
            }

            int stepa3 = bdd.ref(bdd.exists(stepa2, cube));

            int attractorPost_control_state = bdd.ref(stepa3);
            // System.out.println("Attractor with control states under processing in step " + i + ":");
            // bdd.printSet(attractorPost_control_state);

            // [For attractors from uncontrollable locations]
            // Step b1: For ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
            // Step b2: For results in step1, perform AND conjunction with planttransition to derive set of states
            // Step b3: quantify out x_plum, y_plum, z_plum, and these states are the new attractor.
            // Step b4: For these states, we "must" make sure that they do not have any transition which leads regions
            // outside the attractor.
            // Step b5: For S\ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
            // Step b6: For results in step5, perform AND conjunction with planttransition to derive set of states
            // Step b7: quantify out x_plum, y_plum, z_plum
            // Step b8: Perform set difference {Step3}\{Step7}, then states in {Step3}\{Step7} will be guaranteed to
            // have EVERY transition leading to the attractor.

            // int stepb0 = bdd.ref(bdd.exists(attractorPre, cube));
            // int stepb1 = bdd.replace(stepb0, perm);
            int stepb1 = bdd.replace(attractorPre, perm);
            int stepb2 = bdd.and(stepb1, plantTransition);
            int stepb3 = bdd.ref(bdd.exists(stepb2, cube));

            int step = bdd.not(attractorPre); // S\ATTR
            int stepb5_0 = bdd.ref(bdd.exists(step, cube));
            int stepb5 = bdd.replace(stepb5_0, perm);
            int stepb6 = bdd.and(stepb5, plantTransition);
            int stepb7 = bdd.ref(bdd.exists(stepb6, cube));
            int attractorPost_plant_state = bdd.and(stepb3, bdd.not(stepb7));
            // stream.println("Attractor with plant states under processing in step " + i + ":");
            // bdd.printSet(attractorPost_plant_state);

            // Generate the attractor
            int attractorPost = bdd.or(attractorPre, bdd.or(attractorPost_control_state, attractorPost_plant_state));

            // Remove all unused BDD
            // bdd.deref(stepa0);
            bdd.deref(stepa1);
            bdd.deref(stepa2);
            bdd.deref(stepa3);
            bdd.deref(attractorPost_control_state);

            // bdd.deref(stepb0);
            bdd.deref(stepb1);
            bdd.deref(stepb2);
            bdd.deref(stepb3);
            bdd.deref(step);
            bdd.deref(stepb5_0);
            bdd.deref(stepb5);
            bdd.deref(stepb6);
            bdd.deref(stepb7);
            bdd.deref(attractorPost_plant_state);

            // Check if the set of attractors saturates
            if (bdd.ref(bdd.biimp(attractorPre, attractorPost)) == bdd.getOne()) {
                // System.out.println("------ Finish calculating the attractor (winning region) ----------");

                // System.out.println("Do again the controller part to ensure outgoing edge from recur states");
                int stepa0 = bdd.ref(bdd.exists(attractorPre, cube));
                stepa1 = bdd.replace(stepa0, perm);
                stepa2 = bdd.and(stepa1, controllerTransition);


                int stepNew1 = bdd.ref(bdd.and(controllerSynthesisStrategy, stepa2));
                if (bdd.ref(stepNew1) == bdd.getZero()) {
                    System.out.println("Focus: There are something wierd here.");
                }
                int stepNew2 = bdd.ref(bdd.exists(stepNew1, cube)); // K
                int stepNew3 = bdd.ref(bdd.and(controllerSynthesisStrategy, bdd.not(stepNew2)));
                int stepNew4 = bdd.ref(bdd.and(stepa2, bdd.not(stepNew2)));
                bdd.deref(controllerSynthesisStrategy);

                controllerSynthesisStrategy = bdd.or(stepNew3, bdd.or(stepNew4, stepNew1));

                if (bdd.and(attractorPost, initialCondition) != bdd.getZero()) {
                    // System.out.println("Controller exists");
                } else {
                    // System.out.println("Controller does not exist");
                    return bdd.getZero();
                }

                break;
            } else {
                i++;
                bdd.deref(attractorPre); // Dereference the previous image
                attractorPre = bdd.ref(attractorPost); // Assign the post to be the newly generated image
                bdd.deref(attractorPost);
                // System.out.println("-------------------------------------");
            }

        } while (true);

        return controllerSynthesisStrategy;

    }

    /**
     * Find the winning strategy for a safety game.
     * 
     * @param riskRegion the set of risk states
     * @param controllertransition the set of controllable transitions
     * @param planttransition the set of noncontrollable transitions
     * @param perm
     * @param cube
     * @return The set of control transitions which lead to risk states. 
     */
    int findWinningStrategySafety(int riskRegion, int initialCondition, int controllertransition, int planttransition,
            Permutation perm, int cube) {
        // Here we implement the attractor generator: attractor_{0,i}(Reachability) are the set of states which player 0
        // (controller) can force a visit to Reachability in less or equal to i steps.

        int attractorPre = bdd.ref(riskRegion);
        int controllerRiskSynthesisStrategy = bdd.getZero();

        int i = 1;
        do {

            // [For attractors with noncontrollable locations]
            // Step a1: For ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
            // Step a2: perform AND conjunction with controllertransition to derive set of states
            // Step a3: quantify out x_plum, y_plum, z_plum, and these states are the new attractor ATTR'.

            int stepa1 = bdd.replace(attractorPre, perm);
            int stepa2 = bdd.and(stepa1, planttransition);
            int stepa3 = bdd.ref(bdd.exists(stepa2, cube));
            int attractorPost_plant_state = bdd.ref(stepa3);

            // [For attractors from controllable locations]
            // Step b1: For ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
            // Step b2: For results in step1, perform AND conjunction with planttransition to derive set of states
            // Step b3: quantify out x_plum, y_plum, z_plum, and these states are the new attractor.
            // Step b4: For these states, we "must" make sure that they do not have any transition which leads regions
            //          outside the attractor.
            // Step b5: For S\ATTR, perform variable swap from x, y, z to x_plum, y_plum, z_plum
            // Step b6: For results in step5, perform AND conjunction with controllertransition to derive set of states
            // Step b7: quantify out x_plum, y_plum, z_plum
            // Step b8: Perform set difference {Step3}\{Step7}, then states in {Step3}\{Step7} will be guaranteed
            //          to have EVERY transition leading to the attractor.

            int stepb1 = bdd.replace(attractorPre, perm);
            int stepb2 = bdd.and(stepb1, controllertransition);
            int stepb3 = bdd.ref(bdd.exists(stepb2, cube));
            int step = bdd.not(attractorPre); // S\ATTR
            int stepb5_0 = bdd.ref(bdd.exists(step, cube));
            int stepb5 = bdd.replace(stepb5_0, perm);
            int stepb6 = bdd.and(stepb5, controllertransition);
            int stepb7 = bdd.ref(bdd.exists(stepb6, cube));
            int attractorPost_control_state = bdd.and(stepb3, bdd.not(stepb7));

            // Generate the attractor
            int attractorPost = bdd.or(attractorPre, bdd.or(attractorPost_control_state, attractorPost_plant_state));

            // Remove all unused BDD
            bdd.deref(stepa1);
            bdd.deref(stepa2);
            bdd.deref(stepa3);
            bdd.deref(attractorPost_plant_state);

            bdd.deref(stepb1);
            bdd.deref(stepb2);
            bdd.deref(stepb3);
            bdd.deref(step);
            bdd.deref(stepb5_0);
            bdd.deref(stepb5);
            bdd.deref(stepb6);
            bdd.deref(stepb7);
            bdd.deref(attractorPost_control_state);

            // Check if the set of attractors saturates
            if (bdd.ref(bdd.biimp(attractorPre, attractorPost)) == bdd.getOne()) {
                if (bdd.and(attractorPre, initialCondition) != bdd.getZero()) {
                    // Return that all strategies are risky.
                    return bdd.getOne();
                }
                // System.out.println("------ Finish calculating the attractor (losing region) ----------");
                break;

            } else {
                i++;
                bdd.deref(attractorPre); // Dereference the previous image
                // Assign the post to be  the newly generated image
                attractorPre = bdd.ref(attractorPost);
                bdd.deref(attractorPost);

            }

        } while (true);

        // Once the attractor has been decided, we perform a post-processing.
        // Based on the attractor of the set of reachable states, we perform again the reachability analysis.
        // If there exists any thansition which leads to these states, we put them back to the strategy,
        // as we know that these transitions should never be performed.

        int stepa0 = bdd.ref(bdd.exists(attractorPre, cube));
        int stepa1 = bdd.replace(stepa0, perm);
        int stepa2 = bdd.and(stepa1, controllertransition);

        controllerRiskSynthesisStrategy =
                bdd.orTo(controllerRiskSynthesisStrategy, stepa2);

        controllerRiskSynthesisStrategy =
                bdd.orTo(controllerRiskSynthesisStrategy, bdd.and(riskRegion, controllertransition));

        return controllerRiskSynthesisStrategy;
    }

    private MealyMachine generateMealyMachineBuechi(String strategyStringFormat, int initialVertexId, GameArena gameArena) {
        
        
        MealyMachine machine = new MealyMachine();

        if (strategyStringFormat.trim().startsWith("FALSE")) {
            machine.setSolution(false);
            return machine;
        } else {


            String initialID = vertexNameArena.get(initialVertexId).substring(1);

            String[] lineArray = strategyStringFormat.split("[\\r\\n]");

            for (int i = 0; i < lineArray.length; i++) {
                // Retrieve the source, destination, and the memory content
                if (lineArray[i].length() > 1) {

                    StringBuilder sSB = new StringBuilder("");
                    StringBuilder dSB = new StringBuilder("");
                    for (int j = 0; j < lineArray[i].length(); j++) {
                        if (j % 2 == 0) {
                            sSB.append(lineArray[i].substring(j, j + 1));
                        } else {
                            dSB.append(lineArray[i].substring(j, j + 1));
                        }
                    }
                    String source = sSB.toString();
                    String dest = dSB.toString();

                    try {
                        if (source.trim().startsWith("-") || dest.trim().startsWith("-")) {
                            throw new Exception();
                        }
                        int decimalSource = Integer.parseInt(source, 2);
                        int decimalDest = Integer.parseInt(dest, 2);

                        MealyMachineEdgeElement e = new MealyMachineEdgeElement(
                                vertexNameArena.get(decimalSource).split("_")[0].substring(1),
                                vertexNameArena.get(decimalDest).split("_")[0].substring(1),
                                vertexNameArena.get(decimalSource).split("_")[1],
                                getOutputFromBuechiAction(gameArena, decimalSource, decimalDest));

                        machine.getEdges().add(e);
                        machine.getVertices().add(vertexNameArena.get(decimalSource).split("_")[0].substring(1));

                    } catch (Exception ex) {
                        HashSet<String> sourceSet = new HashSet<String>();
                        HashSet<String> destSet = new HashSet<String>();

                        if (source.contains("-")) {
                            sourceSet.addAll(splitVertexFromBDD(source.trim(), 0));
                        } else {
                            sourceSet.add(source);
                        }

                        if (dest.contains("-")) {
                            destSet.addAll(splitVertexFromBDD(dest.trim(), 0));
                        } else {
                            destSet.add(dest);
                        }

                        // Generate all combinations
                        for (String s : sourceSet) {
                            int decimalSource = Integer.parseInt(s, 2);
                            if (decimalSource >= gameArena.vertexList.size()) {
                                continue;
                            }
                            for (String d : destSet) {

                                int decimalDest = Integer.parseInt(d, 2);
                                if (decimalDest >= gameArena.vertexList.size()) {
                                    continue;
                                }

                                machine.getVertices().add(vertexNameArena.get(decimalSource).split("_")[0].substring(1));

                                MealyMachineEdgeElement e = new MealyMachineEdgeElement(
                                        vertexNameArena.get(decimalSource).split("_")[0].substring(1),
                                        vertexNameArena.get(decimalDest).split("_")[0].substring(1),
                                        vertexNameArena.get(decimalSource).split("_")[1],
                                        getOutputFromBuechiAction(gameArena, decimalSource, decimalDest));

                                machine.getEdges().add(e);

                            }
                        }
                    }
                }
            }

            machine.setInitialVertex(initialID);
            machine.setSolution(true);

            return machine;

        }

    }

    private MealyMachine generateMealyMachineSafety(String strategyStringFormat,
            EquivalenceClass initialVertex, ArrayList<EquivalenceClass> safetyArena) {

        MealyMachine machine = new MealyMachine();
        if (strategyStringFormat.trim().startsWith("FALSE")) {
            machine.setSolution(false);
            return machine;
        }

        HashMap<String, String> stateSuccessorStateMap = new HashMap<String, String>();
        HashMap<String, String> stateActionMap = new HashMap<String, String>();

        String[] lineArray = strategyStringFormat.split("[\\r\\n]");
        for (int i = 0; i < lineArray.length; i++) {
            // Retrieve the source, destination, and the memory content
            if (lineArray[i].length() > 1) {

                StringBuilder sSB = new StringBuilder("");
                StringBuilder dSB = new StringBuilder("");
                for (int j = 0; j < lineArray[i].length(); j++) {
                    if (j % 2 == 0) {
                        sSB.append(lineArray[i].substring(j, j + 1));
                    } else {
                        dSB.append(lineArray[i].substring(j, j + 1));
                    }
                }
                String source = sSB.toString();
                String dest = dSB.toString();

                try {
                    if (source.trim().startsWith("-") || dest.trim().startsWith("-")) {
                        throw new Exception();
                    }

                    int decimalSource = Integer.parseInt(source, 2);
                    int decimalDest = Integer.parseInt(dest, 2);

                    if (stateActionMap.get(String.valueOf(decimalSource)) == null) {
                        stateSuccessorStateMap.put(String.valueOf(decimalSource), String.valueOf(decimalDest));
                        stateActionMap.put(String.valueOf(decimalSource), getOutputFromSafetyAction(safetyArena, decimalSource, decimalDest));
                    }

                } catch (Exception ex) {
                    // To extract a string with "-" element, then the conversion requires a recursive call.
                    HashSet<String> sourceSet = new HashSet<String>();
                    HashSet<String> destSet = new HashSet<String>();

                    if (source.contains("-")) {
                        sourceSet.addAll(splitVertexFromBDD(source.trim(), 0));
                    } else {
                        sourceSet.add(source);
                    }

                    if (dest.contains("-")) {
                        destSet.addAll(splitVertexFromBDD(dest.trim(), 0));
                    } else {
                        destSet.add(dest);
                    }

                    // Generate all combinations
                    for (String s : sourceSet) {
                        int decimalSource = Integer.parseInt(s, 2);
                        if (decimalSource >= safetyArena.size()) {
                            continue;
                        }
                        for (String d : destSet) {

                            int decimalDest = Integer.parseInt(d, 2);
                            if (decimalDest >= safetyArena.size()) {
                                continue;
                            }
                            if (stateActionMap.get(String.valueOf(decimalSource)) == null) {
                                stateSuccessorStateMap.put(String.valueOf(decimalSource), String.valueOf(decimalDest));
                                stateActionMap.put(String.valueOf(decimalSource), getOutputFromSafetyAction(safetyArena, decimalSource, decimalDest));
                            }
                        }
                    }
                }
            }
        }

        machine.setSolution(true);

        HashSet<String> envLocations = new HashSet<String>();
        ArrayList<String> workList = new ArrayList<String>();
        workList.add(String.valueOf(initialVertex.id));
        envLocations.add(String.valueOf(initialVertex.id));

        machine.setInitialVertex(String.valueOf(initialVertex.id));

        while (!workList.isEmpty()) {
            String vertexID = workList.remove(0);
            EquivalenceClass v = safetyArena.get(Integer.parseInt(vertexID));
            machine.getVertices().add(String.valueOf(v.id));

            for (String input : v.successor.keySet()) {
                EquivalenceClass succ = v.successor.get(input);
                MealyMachineEdgeElement e = new MealyMachineEdgeElement(String.valueOf(v.id), stateSuccessorStateMap.get(String.valueOf(succ.id)), input, stateActionMap.get(String.valueOf(succ.id)));

                machine.getEdges().add(e);

                if (envLocations.contains(stateSuccessorStateMap.get(String.valueOf(succ.id))) == false) {
                    workList.add(stateSuccessorStateMap.get(String.valueOf(succ.id)));
                    envLocations.add(stateSuccessorStateMap.get(String.valueOf(succ.id)));
                }
            }
        }

        return machine;
    }

    private ArrayList<HashSet> generateCounterStrategySignalsSafety() {
        ArrayList<HashSet> result = new ArrayList<HashSet>();
        return result;
    }

    /**
     * Generate all possible input or output combinations.
     * 
     * @param currentSize
     * @param size
     * @param bitVectors
     * @return 
     */
    private ArrayList<String> generateBitVectors(int currentSize, int size, ArrayList<String> bitVectors) {
        if (currentSize == size) {
            return bitVectors;
        } else {
            ArrayList<String> newBitVectors = new ArrayList<String>();
            for (String bits : bitVectors) {
                newBitVectors.add(bits + "0");
                newBitVectors.add(bits + "1");
            }
            return generateBitVectors(currentSize + 1, size, newBitVectors);
        }
    }

    /**
     * Extract the output from the game-representation Buechi automaton; used 
     * in the  Buechi game solver.
     * 
     * @param arena
     * @param source
     * @param dest
     * @return 
     */
    private String getOutputFromBuechiAction(GameArena arena, int source, int dest) {
        for (VertexEdgeSet v : arena.vertexList) {
            if (v.getVertexID() == source) {
                for (EdgeElement e : v.edgeSet) {
                    if (e.getDestVertexID() == dest) {
                        return e.getEdgeLabel().iterator().next();
                    }
                }
            }
        }
        return "<NO OUTPUT>";
    }

    /**
     * Extract the output from the generated unrolled safety game; used 
     * in the  Co-Buechi + Safety game solver.
     * 
     * @param safetyGame
     * @param source
     * @param dest
     * @return 
     */
    private String getOutputFromSafetyAction(ArrayList<EquivalenceClass> safetyGame, int source, int dest) {
        EquivalenceClass s = safetyGame.get(source);
        EquivalenceClass d = safetyGame.get(dest);
        for (String output : s.successor.keySet()) {
            if (s.successor.get(output) == d) {
                return output;
            }
        }
        return "<NO OUTPUT>";
    }

    /**
     * Generate the controller code for the given LTL specification using the Buechi
     * game solver.
     * 
     * @param prob problem under analysis
     * @param ltl2buechi use LTL2Buchi library as the front-end parser
     * @param outputFormat output format (pseudo, SAL, Ptolemy II)
     * @param proveExistence prove existence or non-existence
     * @return 
     */
    public String invokeMonolithicBuechiEngine(ProblemDescription prob, boolean ltl2buechi, int outputFormat, boolean proveExistence) {

        try {
            GameArena buchiArena = null;
            long startTime = 0;
            long endTime = 0;

            if (ltl2buechi) {
                // Use pure Java-based translator LTL2Buechi
                startTime = System.currentTimeMillis();
                // Step 1: Use LTL2Buchi to generate the corresponding Buechi automaton representation.
                Graph buchiAutomaton;
                if (proveExistence == true) {
                    buchiAutomaton = LTL2Buchi.translate(prob.getLtlSpecification());
                } else {
                    buchiAutomaton = LTL2Buchi.translate("!(" + prob.getLtlSpecification() + ")");
                }


                endTime = System.currentTimeMillis();
                System.out.println("\nTotal elapsed time in execution of method formulaToBA() is : " + (endTime - startTime));

                startTime = System.currentTimeMillis();
                // Step 2: Generate the arena based on reinterpreting the Buechi automata (from the LTL2BA) and the input/output signals.
                buchiArena = createGameArena(prob.getInputSignals(), prob.getOutputSignals(), buchiAutomaton);
                endTime = System.currentTimeMillis();
                System.out.println("Total elapsed time in execution of method createGameArena() is : " + (endTime - startTime));

            }

            startTime = System.currentTimeMillis();
            ArrayList<Integer> finalEnvVertices = new ArrayList<Integer>();
            for (VertexEdgeSet v : buchiArena.vertexList) {
                if (v.getVertexColor() == VertexEdgeSet.COLOR_FINAL) {
                    finalEnvVertices.add(Integer.valueOf(v.getVertexID()));
                }
            }

            MealyMachine machine = analyzeBuechiGame(buchiArena, finalEnvVertices, proveExistence);


            if (proveExistence) {
                if (machine.hasSolution() == true) {
                    // Generate the output format based on the requirement
                    if (outputFormat == OUTPUT_PSUEDO_CODE) {
                        return PseudoCodeTemplate.createPsuedoCode(machine, prob, "Buchi solver");
                    } else if (outputFormat == OUTPUT_SAL) {
                        return SALTemplate.createSALCode(machine, prob, "Buchi solver");
                    } else if (outputFormat == OUTPUT_FSM_ACTOR_PTOLEMY) {
                        ArrayList<String> initialVectorList = new ArrayList<String>();
                        initialVectorList.add("");
                        ArrayList<String> inputBitVectors = generateBitVectors(0, prob.getInputSignals().size(), initialVectorList);
                        return PtolemyTemplate.createPtolemyCode(machine, prob, inputBitVectors);
                    } else {
                        return "Buechi game engine finds the controler, but output format known!";
                    }
                } else {
                    return "Buechi game engine unable to find the controler";
                }
            } else {
                if (machine.hasSolution() == true) {

                    return "Witness of non-existence found by the Buechi game engine";
                } else {
                    return "Buechi game engine unable to find the witness";
                }
            }


        } catch (Exception ex) {
            ex.printStackTrace();
            StringWriter sw = new StringWriter();
            PrintWriter pw = new PrintWriter(sw);
            ex.printStackTrace(pw);
            return sw.toString();
        }
    }

    /**
     * Generate the controller code for the given LTL specification using the 
     * Co-Buechi + safety game solver.
     * 
     * @param prob problem under analysis
     * @param ltl2buechi use LTL2Buchi library as the front-end parser
     * @param outputFormat output format (pseudo, SAL, Ptolemy II)
     * @param proveExistence prove existence or non-existence
     * @return 
     */
    public String invokeMonolithicCoBuechiEngine(ProblemDescription prob,
            boolean ltl2buechi, int outputFormat, boolean proveExistence) {

        try {
            GameArena coBuechiArena = null;
            long startTime = 0;
            long endTime = 0;

            if (ltl2buechi) {
                startTime = System.currentTimeMillis();
                // Step 1: Use LTL2BA to generate the corresponding Buechi automaton representation.
                // Collection<ITransition> buechiAutomatonTransitions = LTL2BA4J.formulaToBA("! (" + prob.getLtlSpecification() + ")");
                Graph coBuechiAutomaton;
                if (proveExistence == true) {
                    coBuechiAutomaton = LTL2Buchi.translate("!(" + prob.getLtlSpecification() + ")");
                } else {
                    coBuechiAutomaton = LTL2Buchi.translate(prob.getLtlSpecification());
                }

                endTime = System.currentTimeMillis();
                System.out.println("\nTotal elapsed time in execution of method formulaToBA() is : " + (endTime - startTime));

                startTime = System.currentTimeMillis();
                // Step 2: Generate the arena based on reinterpreting the Buechi automata (from the LTL2BA) and the input/output signals.
                coBuechiArena = createGameArena(prob.getInputSignals(), prob.getOutputSignals(), coBuechiAutomaton);
                endTime = System.currentTimeMillis();
                System.out.println("Total elapsed time in execution of method createGameArena() is : " + (endTime - startTime));

            }

            // Step 3: Generate risk states
            ArrayList<Integer> riskStates = new ArrayList<Integer>();
            String initialVertexID = "";
            for (VertexEdgeSet v : coBuechiArena.vertexList) {
                if (v.getVertexColor() == VertexEdgeSet.COLOR_FINAL) {
                    riskStates.add(Integer.valueOf(v.getVertexID()));
                }
                if (v.getVertexProperty() == VertexEdgeSet.INITIAL_PLANT) {
                    initialVertexID = String.valueOf(v.getVertexID());
                }
            }

            ArrayList<String> initialVectorList = new ArrayList<String>();
            initialVectorList.add("");
            ArrayList<String> inputBitVectors = generateBitVectors(0, prob.getInputSignals().size(), initialVectorList);
            initialVectorList = new ArrayList<String>();
            initialVectorList.add("");
            ArrayList<String> outputBitVectors = generateBitVectors(0, prob.getOutputSignals().size(), initialVectorList);

            // Step 4: Invoke safety game translation 
            startTime = System.currentTimeMillis();
            CoBuechiSafetyReduction reduction = new CoBuechiSafetyReduction(coBuechiArena, riskStates, inputBitVectors.size(), inputBitVectors, outputBitVectors);
            ArrayList<EquivalenceClass> safetyGameArena = reduction.createReductionGraph(initialVertexID, 1,
                    prob.getUnrollSteps() * 2 + 1, MAX_VISIT_COBUECHI_FINAL_STATE, inputBitVectors, outputBitVectors);
            endTime = System.currentTimeMillis();
            System.out.println("Total elapsed time in execution of method createReductionGraph() is: " + (endTime - startTime));

            // Step 5: Execute the safety game engine. 
            startTime = System.currentTimeMillis();
            MealyMachine machine = analyzeSafetyGameFromCoBuechi(safetyGameArena, reduction.initialVertex, reduction.riskVertex, proveExistence);

            endTime = System.currentTimeMillis();
            System.out.println("Total elapsed time in execution of method analyzeSafetyGame() is: " + (endTime - startTime));

            if (proveExistence) {
                if (machine.hasSolution() == true) {
                    // Generate the output format based on the requirement
                    if (outputFormat == OUTPUT_SAL) {
                        return SALTemplate.createSALCode(machine, prob, "CoBuchi+safety solver");
                    } else if (outputFormat == OUTPUT_PSUEDO_CODE) {
                        return PseudoCodeTemplate.createPsuedoCode(machine, prob, "CoBuchi+safety solver");
                    } else if (outputFormat == OUTPUT_FSM_ACTOR_PTOLEMY) {
                        // return PtolemyTemplate.createPtolemyCode(machine, prob);
                        return PtolemyTemplate.createPtolemyCode(machine, prob, inputBitVectors);
                    } else {
                        return "Buechi game engine finds the controler, but output format known!";
                    }
                } else {
                    return "Co-Buechi + safety game engine unable to find the controler";
                }
            } else {
                if (machine.hasSolution() == true) {
                    return "Witness of non-existence found by the Co-Buechi + safety game engine";
                } else {
                    return "Co-Buechi + safety game engine unable to find the witness";
                }
            }

        } catch (Exception ex) {
            ex.printStackTrace();
            StringWriter sw = new StringWriter();
            PrintWriter pw = new PrintWriter(sw);
            ex.printStackTrace(pw);
            return sw.toString();
        }
    }

    /**
     * Invoke the synthesis engine (without a call from the GUI).
     * 
     * @param prob
     * @param isBuechiSolver
     * @return 
     */
    public MealyMachine invokeSynthesisLibrary(ProblemDescription prob, boolean isBuechiSolver) {
        try {
            Graph buchiAutomaton = LTL2Buchi.translate(prob.getLtlSpecification());
            GameArena buchiArena = createGameArena(prob.getInputSignals(), prob.getOutputSignals(), buchiAutomaton);
            if (isBuechiSolver) {
                ArrayList<Integer> finalEnvVertices = new ArrayList<Integer>();
                for (VertexEdgeSet v : buchiArena.vertexList) {
                    if (v.getVertexColor() == VertexEdgeSet.COLOR_FINAL) {
                        finalEnvVertices.add(Integer.valueOf(v.getVertexID()));
                    }
                }
                return analyzeBuechiGame(buchiArena, finalEnvVertices, true);

            } else {

                // Step 1: Use LTL2BA to generate the corresponding Buechi automaton representation.
                // Collection<ITransition> buechiAutomatonTransitions = LTL2BA4J.formulaToBA("! (" + prob.getLtlSpecification() + ")");
                Graph coBuechiAutomaton = LTL2Buchi.translate("!(" + prob.getLtlSpecification() + ")");

                // Step 2: Generate the arena based on reinterpreting the Buechi automata (from the LTL2BA) and the input/output signals.
                GameArena coBuechiArena = createGameArena(prob.getInputSignals(), prob.getOutputSignals(), coBuechiAutomaton);

                // Step 3: Risk states
                ArrayList<Integer> riskStates = new ArrayList<Integer>();
                String initialVertexID = "";
                for (VertexEdgeSet v : coBuechiArena.vertexList) {
                    if (v.getVertexColor() == VertexEdgeSet.COLOR_FINAL) {
                        riskStates.add(Integer.valueOf(v.getVertexID()));
                    }
                    if (v.getVertexProperty() == VertexEdgeSet.INITIAL_PLANT) {
                        initialVertexID = String.valueOf(v.getVertexID());
                    }
                }

                ArrayList<String> initialVectorList = new ArrayList<String>();
                initialVectorList.add("");
                ArrayList<String> inputBitVectors = generateBitVectors(0, prob.getInputSignals().size(), initialVectorList);
                initialVectorList = new ArrayList<String>();
                initialVectorList.add("");
                ArrayList<String> outputBitVectors = generateBitVectors(0, prob.getOutputSignals().size(), initialVectorList);

                // Step 4: Invoke safety game translation 
                CoBuechiSafetyReduction reduction = new CoBuechiSafetyReduction(coBuechiArena, riskStates, inputBitVectors.size(), inputBitVectors, outputBitVectors);
                ArrayList<EquivalenceClass> safetyGameArena = reduction.createReductionGraph(initialVertexID, 1,
                        prob.getUnrollSteps() * 2 + 1, MAX_VISIT_COBUECHI_FINAL_STATE, inputBitVectors, outputBitVectors);

                // Step 5: Execute the safety game engine. 
                return analyzeSafetyGameFromCoBuechi(safetyGameArena, reduction.initialVertex, reduction.riskVertex, true);
            }
        } catch (Exception ex) {
            StringWriter sw = new StringWriter();
            PrintWriter pw = new PrintWriter(sw);
            ex.printStackTrace(pw);
            return null;
        }
    }

    /**
     * Padd 0 before a string to a specified length.
     * 
     * @param s
     * @param idealLength
     * @return 
     */
    private static String padZeroToString(String s, int specifiedLength) {
        if (s.length() > specifiedLength) {
            System.out.println("Error: the string under padding is longer than the ideal length");
            return s;
        } else {
            StringBuffer zeroString = new StringBuffer("");
            for (int i = 0; i < specifiedLength - s.length(); i++) {
                zeroString.append("0");
            }
            return (zeroString.append(s)).toString();
        }

    }

    private int post(int i) {
        return 2 * i + 1;
    }

    private int pre(int i) {
        return 2 * i;
    }

    /**
     * Print the textural presentation of a Buechi automata; debug only.
     * 
     * @param spec 
     */
    private void printBuechiAutomata(String spec) {

        try {
            Graph graph = LTL2Buchi.translate(spec);
            Node init = graph.getInit();

            List nodes = graph.getNodes();
            for (Iterator i = nodes.iterator(); i.hasNext();) {
                Node n = (Node) i.next();
                if (init == n) {
                    if (init.getBooleanAttribute("accepting")) {
                        System.out.println("State (init, acc) " + init.getId());
                    } else {
                        System.out.println("State (init) " + init.getId());
                    }
                } else {
                    if (n.getBooleanAttribute("accepting")) {
                        System.out.println("State (acc) " + n.getId());
                    } else {
                        System.out.println("State " + n.getId());
                    }
                }
                for (Iterator j = n.getOutgoingEdges().iterator(); j.hasNext();) {

                    Edge e = (Edge) j.next();
                    System.out.println("\t (" + e.getGuard() + ") -> " + e.getNext().getId());
                    String guard = e.getGuard().equals("-") ? "1" : e.getGuard();

                    StringTokenizer tok = new StringTokenizer(new String(guard), "&");
                    guard = "";

                    while (tok.hasMoreTokens()) {
                        guard += tok.nextToken();

                        if (tok.hasMoreTokens()) {
                            guard += " && ";
                        }
                    }

                    tok = new StringTokenizer(new String(guard), "|");
                    guard = "";

                    while (tok.hasMoreTokens()) {
                        guard += tok.nextToken();

                        if (tok.hasMoreTokens()) {
                            guard += " || ";
                        }
                    }
                }
            }

        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Analyze the string to know how many vertices are encoded.
     * 
     * @param underProcessing
     * @param lengthIndex
     * @return
     */
    private HashSet<String> splitVertexFromBDD(String underProcessing, int lengthIndex) {
        if (lengthIndex == underProcessing.length()) {
            HashSet<String> returnSet = new HashSet<String>();
            returnSet.add("");
            return returnSet;
        } else {
            if (underProcessing.substring(lengthIndex, lengthIndex + 1).equalsIgnoreCase("-")) {
                HashSet<String> returnSet = new HashSet<String>();
                lengthIndex++;

                HashSet<String> subStringSet = splitVertexFromBDD(underProcessing, lengthIndex);
                for (String str : subStringSet) {
                    returnSet.add("0" + str);
                    returnSet.add("1" + str);
                }

                return returnSet;
            } else {
                HashSet<String> returnSet = new HashSet<String>();
                HashSet<String> subStringSet = splitVertexFromBDD(underProcessing, lengthIndex + 1);
                for (String str : subStringSet) {
                    returnSet.add(underProcessing.substring(lengthIndex, lengthIndex + 1) + str);
                }
                return returnSet;
            }
        }
    }
}
